from IPython.display import display,clear_output 
class ChatGLM(object):
    def __init__(self,
                 model,
                 tokenizer,
                 stream=True,
                 max_chat_rounds=20,
                 history=None,
                 max_length=8192,
                 num_beams=1,
                 do_sample=True,
                 top_p=0.8,
                 temperature=0.8,
                 logits_processor=None
                ):
        
        self.__dict__.update(locals())
        self.history =  [] if history is None else history
        
        try:
            self.register_magic() 
            print('register magic %%chatglm sucessed ...')
        except Exception as err:
            print('register magic %%chatglm failed ...')
            print(err)
        
        response = self('你好')
        if not self.stream:
            print(response)
        self.history = self.history[:-1]

    
    def __call__(self,query):
        len_his = len(self.history)
        if len_his>=self.max_chat_rounds+1:
            self.history = self.history[len_his-self.max_chat_rounds:]
        
        if not self.stream:
            response,self.history  = self.model.chat(self.tokenizer,
             query,self.history,self.max_length,self.num_beams,
             self.do_sample,self.top_p,self.temperature,
             self.logits_processor)
            return response 
        
        result = self.model.stream_chat(self.tokenizer,
            query,self.history,None,self.max_length,
            self.do_sample,self.top_p,self.temperature,
            self.logits_processor,None)
        
        for response,history in result:
            print(response)
            clear_output(wait=True)
        self.history = history
        #clear_output(waite=True)
        return response
    
    def register_magic(self):
        import IPython
        from IPython.core.magic import (Magics, magics_class, line_magic,
                                        cell_magic, line_cell_magic)
        @magics_class
        class ChatGLMMagics(Magics):
            def __init__(self,shell, pipe):
                super().__init__(shell)
                self.pipe = pipe

            @line_cell_magic
            def chatglm(self, line, cell=None):
                "Magic that works both as %chatglm and as %%chatglm"
                if cell is None:
                    return self.pipe(line)
                else:
                    print(self.pipe(cell))
                    
        ipython = IPython.get_ipython()
        magic = ChatGLMMagics(ipython,self)
        ipython.register_magics(magic)
        


import pandas as pd 
import numpy as np 
from IPython.display  import display 

#Visual Metrics in Notebook for Catboost
class VisCallback:
    def __init__(self, figsize = (8,6), metric='val_Accuracy', mode='max'):
        import matplotlib.pyplot as plt
        self.plt = plt
        self.figsize = figsize
        self.metric = metric
        self.mode = mode
        x_bounds = [0, 10]
        title = f'best {metric} = ?'
        self.update_graph(None, title=title, x_bounds = x_bounds)
        
    def get_dfhistory(self, info):
        from copy import deepcopy
        if not hasattr(info,'metrics'):
            return pd.DataFrame() 
        
        dic = deepcopy(info.metrics) 
        if 'learn' in dic:
            dic['train'] = dic['learn']
            dic.pop('learn')

        if 'validation' in dic:
            dic['val'] = dic['validation']
            dic.pop('validation')

        dfhis_train = pd.DataFrame(dic['train']) 
        dfhis_train.columns = ['train_'+x for x in dfhis_train.columns]

        dfhis_val = pd.DataFrame(dic['val']) 
        dfhis_val.columns = ['val_'+x for x in dfhis_val.columns]
        dfhistory = dfhis_train.join(dfhis_val)
        dfhistory['iteration'] = range(1,len(dfhistory)+1)
        
        return dfhistory 

    def after_iteration(self, info):
        dfhistory = self.get_dfhistory(info)
        n = len(dfhistory)
        x_bounds = [dfhistory['iteration'].min(), 10+(n//10)*10]
        title = self.get_title(info)
        self.update_graph(info, title = title, x_bounds = x_bounds)
        return True
        
    def get_best_score(self, info):
        dfhistory = self.get_dfhistory(info)
        arr_scores = dfhistory[self.metric]
        best_score = np.max(arr_scores) if self.mode=="max" else np.min(arr_scores)
        best_iteration = dfhistory.loc[arr_scores==best_score,'iteration'].tolist()[0]
        return (best_iteration, best_score)
        
    def get_title(self,  info):
        best_iteration,best_score = self.get_best_score(info)
        title = f'best {self.metric} = {best_score:.4f} (@iteration {best_iteration})'
        return title

    def update_graph(self, info, title=None, x_bounds=None, y_bounds=None):
        if not hasattr(self, 'graph_fig'):
            self.graph_fig, self.graph_ax = self.plt.subplots(1, figsize=self.figsize)
            self.graph_out = display(self.graph_ax.figure, display_id=True)
        self.graph_ax.clear()
        dfhistory = self.get_dfhistory(info)
        iterations = dfhistory['iteration'] if 'iteration' in dfhistory.columns else []
        
        m1 = "train_"+self.metric.replace('val_','')
        if  m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            
            self.graph_ax.plot(iterations,train_metrics,'bo--',label= m1)

        m2 = 'val_'+self.metric.replace('val_','')
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.graph_ax.plot(iterations,val_metrics,'co-',label =m2)


        self.graph_ax.set_xlabel("iteration")
        self.graph_ax.set_ylabel(self.metric.replace('val_',''))  
        if title:
             self.graph_ax.set_title(title)
        if m1 in dfhistory.columns or m2 in dfhistory.columns or self.metric in dfhistory.columns:
            self.graph_ax.legend(loc='best')
            
        if len(iterations)>0:
            best_iteration, best_score = self.get_best_score(info)
            self.graph_ax.plot(best_iteration,best_score,'r*',markersize=15)

        if x_bounds is not None: self.graph_ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.graph_ax.set_ylim(*y_bounds)
        self.graph_out.update(self.graph_ax.figure)
        self.plt.close();
        